#ifndef __GEOMETRY_DEFINITIONS_H__
#define __GEOMETRY_DEFINITIONS_H__

/*
 * GEOMETRY_DEFINITIONS.H
 */
 
#ifndef __DEFINITIONS_H__
#include "definitions.h"
#endif

/* ---------- constants */

enum
{
	GEOMETRY_GROUP_TAG= 0x67656F6D,	// 'geom' looks like text
	GEOMETRY_VERSION_NUMBER= 1,
	
	MAXIMUM_MATERIALS_PER_GEOMETRY= 32,
	MAXIMUM_SURFACES_PER_GEOMETRY= 512,
	MAXIMUM_VERTICES_PER_GEOMETRY= 2048,

	GEOMETRY_MINIMUM_SCALE= (WORLD_ONE/5),
	GEOMETRY_NORMAL_SCALE= WORLD_ONE,
	GEOMETRY_MAXIMUM_SCALE= (3*WORLD_ONE)
};

/* ---------- structures */

/* ---------- geometry definition */

#define SIZEOF_STRUCT_GEOMETRY_VERTEX 6
struct geometry_vertex
{
	short_world_point3d p;
};

enum
{
	NUMBER_OF_CORNERS_PER_TRIANGLE= 3,
	NUMBER_OF_ENCODED_GEOMETRY_CORNER_NORMALS= 1024,
	NUMBER_OF_GEOMETRY_CORNER_INTERPOLANTS= 2
};

#define SIZEOF_STRUCT_GEOMETRY_CORNER 8
struct geometry_corner
{
	fixed_fraction interpolants[NUMBER_OF_GEOMETRY_CORNER_INTERPOLANTS];
	
	short vertex_index;
	short encoded_normal; // used for normal-based shading (environment mapping, etc.)
};

enum // struct geometry_surface flags
{
	_geometry_surface_is_transparent_bit,	// draw back-to-front
	_geometry_surface_no_backface_cull_bit,	// don't cull; used for glass, etc.
	_geometry_surface_under_mesh_bit,	// its just water under the bridge
	_geometry_surface_reflective_bit, // has a reflective surface attached
	_geometry_surface_uses_closest_z_bit, // use closest z for rendering (instead of farthest)

	// these bits are set and reset by the model for rendering, collision detection, etc..
	_geometry_surface_negative_bit,	// for iterative traversal-> used at runtime
	_geometry_surface_does_not_stop_projectiles_bit,	// for iterative traversal-> used at runtime
	
	NUMBER_OF_GEOMETRY_SURFACE_FLAGS, // <=16
	
	_geometry_surface_is_transparent_flag= FLAG(_geometry_surface_is_transparent_bit),
	_geometry_surface_no_backface_cull_flag= FLAG(_geometry_surface_no_backface_cull_bit),
	_geometry_surface_under_mesh_flag= FLAG(_geometry_surface_under_mesh_bit),
	_geometry_surface_reflective_flag= FLAG(_geometry_surface_reflective_bit),
	_geometry_surface_uses_closest_z_flag= FLAG(_geometry_surface_uses_closest_z_bit), // use closest z for rendering (instead of farthest)
	
	_geometry_surface_negative_flag= FLAG(_geometry_surface_negative_bit),
	_geometry_surface_does_not_stop_projectiles_flag= FLAG(_geometry_surface_does_not_stop_projectiles_bit)
};

#define BUILD_MESH_CELL_CONTAINMENT(x, y) (((y)<<8)|(x))
#define GET_MESH_CELL_CONTAINMENT_X(c) ((c)&0xff)
#define GET_MESH_CELL_CONTAINMENT_Y(c) (((c)>>8)&0xff)

#define SIZEOF_STRUCT_GEOMETRY_SURFACE 64
struct geometry_surface
{
	word flags;
	
	world_vector3d normal;
	world_distance d;				// of plane equation

	short_world_rectangle3d bounds;	// bounding box in geometry space for this triangle
	
	fixed_fraction transparency; // was transparency
	short material_index;
	
	struct geometry_corner corners[NUMBER_OF_CORNERS_PER_TRIANGLE];

	// bsp tree shit
	short next_negative_surface_index;
	short next_parallel_surface_index;
	short next_positive_surface_index;

	word obsolete; //mesh_cell_contained;
	short dependency_table_index;
	short unused[1];
};

enum
{
	MAXIMUM_GEOMETRY_MATERIAL_NAME_LENGTH= 32
};

#define SIZEOF_STRUCT_GEOMETRY_MATERIAL 64
struct geometry_material
{
	char name[MAXIMUM_GEOMETRY_MATERIAL_NAME_LENGTH];	// used only in fear
	short sequence_index;

	short unused[12];
	
	// these two fields are filled in before rendering
	short collection_index;
	short color_table_index;
	short bitmap_index;
};

enum // struct geometry_definition flags (also user_data contents)
{
	_geometry_definition_is_mesh_aligned_bit,
	_geometry_definition_is_object_bit, // completely fits within one cell...
	
	NUMBER_OF_GEOMETRY_DEFINITION_FLAGS, // <=31
	
	_geometry_definition_is_mesh_aligned_flag= FLAG(_geometry_definition_is_mesh_aligned_bit),
	_geometry_definition_is_object_flag= FLAG(_geometry_definition_is_object_bit)
};

#define SIZEOF_STRUCT_GEOMETRY_DEFINITION 128
struct geometry_definition
{
	unsigned long flags;
	
	file_tag collection_reference_tag;
	
	short material_count;
	short vertex_count;
	short surface_count;
	short dependency_count;

	short_world_point3d center; // in geometry coordinate space.
	short height;
	long obsolete;
	
	int materials_offset, materials_size;
	struct geometry_material *materials;	// memory-based
	
	int vertices_offset, vertices_size;
	struct geometry_vertex *vertices; // memory-based
	
	int surfaces_offset, surfaces_size;
	struct geometry_surface *surfaces; // memory-based

	int dependency_list_offset, dependency_size;
	short *dependency_list;
	
	long data_offset, data_size;
	void *data; // memory-based
	
	short_world_rectangle3d bounds;

	short unused[13];

	short collection_reference_index; // only valid at runtime..
};

/* ---------- prototypes */
#define initialize_geometry_definitions() \
	initialize_definitions(&geometries)
#define initialize_geometry_definitions_for_new_map() \
	initialize_definitions_for_new_map(&geometries)
#define geometry_definition_tag_to_geometry_type(subgroup) \
	definition_tag_to_index(&geometries, (subgroup))

void dispose_geometry_definitions_from_old_map(void);

#define get_geometry_definition_tag_header(index) get_definition_tag_header(&geometries, (index))

// because collision detection (models.c) and drawing need this (render_models.c)
void transform_geometry_point_to_world(world_point3d *origin, short_world_point3d *center,
	short_world_point3d *p, matrix3x3 *geometry_to_world_matrix, world_point3d *result);
void transform_world_point_to_geometry(world_point3d *origin, short_world_point3d *center,
	world_point3d *p, matrix3x3 *world_to_geometry_matrix, short_world_point3d *result);
void get_geometry_matrices(world_vector3d *rotation_vector,
	world_vector3d *up_vector,
	world_vector3d *scale_vector, // (WORLD_ONE/5, 3*WORLD_ONE)
	matrix3x3 *geometry_to_world_matrix, matrix3x3 *world_to_geometry_matrix);
world_vector3d *rotate_model_vector(world_vector3d *result, matrix3x3 *rotation,
	world_vector3d *source);

// probably should be in world.c? (only geometries use this though)
void downshift_trig_matrix_into_world_matrix(
	matrix3x3 *m);

void calculate_geometry_cell_width_and_height(struct geometry_definition *geometry,
	short *mesh_cell_width, short *mesh_cell_height);

boolean load_geometry_data(long tag_index, struct geometry_definition *definition, boolean readahead);

#ifdef DEBUG
#define get_geometry_definition(index) ((struct geometry_definition *)get_definition(&geometries, (index)))
struct geometry_vertex *get_geometry_vertex(struct geometry_definition *geometry, int index);
struct geometry_surface *get_geometry_surface(struct geometry_definition *geometry, int surface_index);
struct geometry_material *get_geometry_material(struct geometry_definition *geometry, int material_index);
#else
#define get_geometry_definition(index) ((struct geometry_definition *)(geometries.definitions)+(index))
#define get_geometry_vertex(geometry, index) ((geometry)->vertices+(index))
#define get_geometry_surface(geometry, index) ((geometry)->surfaces+(index))
#define get_geometry_material(geometry, index) ((geometry)->materials+(index))
#endif

/* ---------- globals */

extern struct definition_globals geometries;

#endif
